@@ -31,7 +31,11 @@
 import sys
 import posixpath
 import collections
-import distutils.version
+
+try:
+    from pkg_resources import parse_version
+except:
+    from distutils.version import StrictVersion as parse_version
 
 import numpy as np
 import h5py
@@ -174,7 +178,7 @@ def read_object_array(f, data, options):
 
     Returns
     -------
-    numpy.ndarray of numpy.object_
+    numpy.ndarray of numpy.object\\_
         The Python object array containing the items pointed to by
         `data`.
 
@@ -485,6 +489,7 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
                       np.float32, np.float64,
                       np.complex64, np.complex128,
                       np.bytes_, np.unicode_, np.object_]
+        self._numpy_types = list(self.types)
         # Using Python 3 type strings.
         self.python_type_strings = ['numpy.ndarray', 'numpy.matrix',
                                     'numpy.chararray',
@@ -546,8 +551,8 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
         self.matlab_classes = list(self.__MATLAB_classes.values())
 
         # For h5py >= 2.2, half precisions (np.float16) are supported.
-        if distutils.version.LooseVersion(_H5PY_VERSION) \
-                >= distutils.version.LooseVersion('2.2'):
+        if parse_version(_H5PY_VERSION) \
+                >= parse_version('2.2'):
             self.types.append(np.float16)
             self.python_type_strings.append('numpy.float16')
 
@@ -689,10 +694,9 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
                 is not h5py.Reference \
                 and not np.iscomplexobj(data) \
                 and (options.structured_numpy_ndarray_as_struct \
-                or ", 'O'" in str(data_to_store.dtype) \
-                or not all(data_to_store.shape) \
-                or not all([all(data_to_store[n].shape) \
-                for n in data_to_store.dtype.names])):
+                or (data_to_store.dtype.hasobject \
+                or '\\x00' in str(data_to_store.dtype)) \
+                or does_dtype_have_a_zero_shape(data_to_store.dtype)):
             wrote_as_struct = True
             # Grab the list of fields that don't have a null character
             # or a / in them since those can't be written.
@@ -928,9 +932,9 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
             # written as a vlen='S1' array of bytes_ arrays of the
             # individual characters.
             if options.matlab_compatible \
-                    and distutils.version.LooseVersion( \
+                    and parse_version( \
                     _H5PY_VERSION) \
-                    >= distutils.version.LooseVersion('2.3'):
+                    >= parse_version('2.3'):
                 try:
                     dt = h5py.special_dtype(vlen=np.dtype('S1'))
                     fs = np.empty(shape=(len(field_names),), dtype=dt)
@@ -1013,8 +1017,8 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
             defaultfactory = type(None)
         else:
             defaultfactory = lambda : None
-        attributes = collections.defaultdict(defaultfactory,
-                                             grp[name].attrs.items())
+        attributes = collections.defaultdict(defaultfactory)
+        read_all_attributes_into(grp[name].attrs, attributes)
 
         str_attrs = dict()
         for attr_name in ('Python.Type',
@@ -1052,12 +1056,8 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
             python_fields = [convert_to_str(x)
                              for x in python_fields]
 
-        # If we are using h5py version >= 2.3, we can actually read the
-        # MATLAB_fields Attribute if it is present.
-        matlab_fields = None
-        if distutils.version.LooseVersion(_H5PY_VERSION) \
-                >= distutils.version.LooseVersion('2.3'):
-            matlab_fields = attributes['MATLAB_fields']
+        # Read the MATLAB_fields Attribute if it was present.
+        matlab_fields = attributes['MATLAB_fields']
 
         # If it is a Dataset, it can simply be read and then acted upon
         # (if it is an HDF5 Reference array, it will need to be read
@@ -1130,7 +1130,7 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
                 if python_fields is not None:
                     fields = python_fields
                 else:
-                    fields = [k.tostring().decode()
+                    fields = [numpy_to_bytes(k).decode()
                               for k in matlab_fields]
                 # Now, there may be fields available that were not
                 # given, but still should be read. Keys that are not in
@@ -1157,8 +1157,8 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
                 # If any of the elements are not Numpy types or if they
                 # don't all have the exact same dtype and shape, then
                 # this field will just be an object field.
-                if v.size == 0 or not isinstance(v.flatten()[0], \
-                        tuple(self.types)):
+                if v.size == 0 or type(v.flat[0]) \
+                        not in self._numpy_types:
                     dt_whole.append((k_name, 'object'))
                     continue
 
@@ -1218,7 +1218,7 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
                 if python_fields is not None:
                     fields = python_fields
                 else:
-                    fields = [k.tostring().decode()
+                    fields = [numpy_to_bytes(k).decode()
                               for k in matlab_fields]
                 struct_dtype = list()
                 for k in fields:
@@ -1363,10 +1363,11 @@ class NumpyScalarArrayMarshaller(TypeMarshaller):
                     dt_whole = list()
                     for k in matlab_fields:
                         if sys.hexversion >= 0x03000000:
-                            dt_whole.append((k.tostring().decode(),
+                            dt_whole.append((numpy_to_bytes(k).decode(),
                                             'object'))
                         else:
-                            dt_whole.append((k.tostring(), 'object'))
+                            dt_whole.append((numpy_to_bytes(k),
+                                             'object'))
                     data = np.zeros(shape=tuple(np.uint64(data)),
                                     dtype=dt_whole)
 
@@ -1654,8 +1655,8 @@ class PythonDictMarshaller(TypeMarshaller):
         # should be deleted. It is written as a vlen='S1' array of
         # bytes_ arrays of the individual characters.
         if options.matlab_compatible \
-                and distutils.version.LooseVersion(_H5PY_VERSION) \
-                >= distutils.version.LooseVersion('2.3'):
+                and parse_version(_H5PY_VERSION) \
+                >= parse_version('2.3'):
             try:
                 dt = h5py.special_dtype(vlen=np.dtype('S1'))
                 fs = np.empty(shape=(len(fields),), dtype=dt)
