@@ -29,13 +29,211 @@
 """
 
 import sys
+import collections
 import copy
 import random
 
+# In old versions of Python, everyting in collections.abc was in
+# collections.
+try:
+    import collections.abc
+except ImportError:
+    collections.abc = collections
+
+try:
+    from pkg_resources import parse_version
+except:
+    from distutils.version import StrictVersion as parse_version\
+
 import numpy as np
 import h5py
 
 
+# We need to determine if h5py is one of the versions that cannot read
+# the MATLAB_fields Attribute in the normal fashion so that we can
+# handle it specially.
+_cant_read_matlab_fields = (
+    parse_version(h5py.__version__)
+    < parse_version('2.3'))
+_handle_matlab_fields_specially = (
+    parse_version(h5py.__version__)
+    in (parse_version('3.0'),
+        parse_version('3.1')))
+
+if _handle_matlab_fields_specially:
+    import ctypes
+    import h5py._objects
+
+
+# Determine if numpy.ndarrays and scalars have a tobytes method or
+# not. If they don't, then the tostring method must be used instead.
+_numpy_has_tobytes = hasattr(np.array([1]), 'tobytes')
+
+
+def numpy_to_bytes(obj):
+    """ Get the raw bytes of a numpy object's data.
+
+    Calls the ``tobytes`` method on `obj` for new versions of ``numpy``
+    where the method exists, and ``tostring`` for old versions of
+    ``numpy`` where it does not.
+
+    Parameters
+    ----------
+    obj : numpy.generic or numpy.ndarray
+        Numpy scalar or array.
+
+    Returns
+    -------
+    data : bytes
+        The raw data.
+
+    """
+    if _numpy_has_tobytes:
+        return obj.tobytes()
+    else:
+        return obj.tostring()
+
+
+def read_matlab_fields_attribute(attrs):
+    """ Reads the ``MATLAB_fields`` Attribute.
+
+    On some versions of ``h5py``, the ``MATLAB_fields`` Attribute cannot
+    be read in the standard way and must instead be read in a more
+    manual fashion. This function reads the Attribute by the proper
+    method.
+
+    Parameters
+    ----------
+    attrs : h5py.AttributeManager
+        The Attribute manager to read from.
+
+    Returns
+    -------
+    value : numpy.ndarray or None
+        The value of the ``MATLAB_fields`` Attribute, or ``None`` if it
+        isn't available or its format is invalid.
+
+    Raises
+    ------
+    TypeError
+        If an argument has the wrong type.
+
+    """
+    if not isinstance(attrs, h5py.AttributeManager):
+        raise TypeError('attrs must be a h5py.AttributeManager.')
+    if not _handle_matlab_fields_specially:
+        return attrs.get('MATLAB_fields')
+    if _cant_read_matlab_fields or 'MATLAB_fields' not in attrs:
+        return None
+    # The following method is loosely based on the method provided by
+    # takluyver at
+    # https://github.com/h5py/h5py/issues/1817#issuecomment-781385699
+    #
+    # but has been improved by reading it directly as (size_t, void*)
+    # pairs uint64s instead of using struct.unpack, and avoiding making
+    # intermediate copies of the data by copying directly to the output
+    # array.
+    with h5py._objects.phil:
+        attr_id = attrs.get_id('MATLAB_fields')
+        dt = np.dtype([('length', np.uintp), ('pointer', np.intp)])
+        raw_buf = np.empty(attr_id.shape, dtype=dt)
+        attr_id.read(raw_buf, mtype=attr_id.get_type())
+        attr = np.empty(raw_buf.shape, dtype='object')
+        for i, (length, ptr) in enumerate(raw_buf.flat):
+            at = np.empty(length, dtype='S1')
+            ctypes.memmove(at.ctypes.data, int(ptr), int(length))
+            attr.flat[i] = at
+        return attr
+
+
+def read_all_attributes_into(attrs, out):
+    """ Reads all Attributes into a MutableMapping (dict-like)
+
+    Reads all Attributes into the MutableMapping (dict-like) out,
+    including the special handling of the ``MATLAB_fields`` Attribute on
+    versions of ``h5py`` where it cannot be read in the standard
+    fashion.
+
+    Parameters
+    ----------
+    attrs : h5py.AttributeManager
+        The Attribute manager to read from.
+    out : MutableMapping
+        The MutableMapping (dict-like) to write the Attributes into.
+
+    Raises
+    ------
+    TypeError
+        If an argument has the wrong type.
+
+    See Also
+    --------
+    read_matlab_fields_attribute
+
+    """
+    if not isinstance(attrs, h5py.AttributeManager):
+        raise TypeError('attrs must be a h5py.AttributeManager.')
+    if not isinstance(out, (dict, collections.defaultdict,
+                            collections.abc.MutableMapping)):
+        raise TypeError('out must be a MutableMapping.')
+    if not _handle_matlab_fields_specially \
+            or 'MATLAB_fields' not in attrs:
+        out.update(attrs.items())
+    else:
+        for k in attrs:
+            if k != 'MATLAB_fields':
+                out[k] = attrs[k]
+            else:
+                out[k] = read_matlab_fields_attribute(attrs)
+
+
+def does_dtype_have_a_zero_shape(dt):
+    """ Determine whether a dtype (or its fields) have zero shape.
+
+    Determines whether the given ``numpy.dtype`` has a shape with a zero
+    element or if one of its fields does, or if one of its fields'
+    fields does, and so on recursively. The following dtypes do not have
+    zero shape.
+
+    * ``'uint8'``
+    * ``[('a', 'int32'), ('blah', 'float16', (3, 3))]``
+    * ``[('a', [('b', 'complex64')], (2, 1, 3))]``
+
+    But the following do
+
+    * ``('uint8', (1, 0))``
+    * ``[('a', 'int32'), ('blah', 'float16', (3, 0))]``
+    * ``[('a', [('b', 'complex64')], (2, 0, 3))]``
+
+    Parameters
+    ----------
+    dt : numpy.dtype
+        The dtype to check.
+
+    Returns
+    -------
+    yesno : bool
+        Whether `dt` or one of its fields has a shape with at least one
+        element that is zero.
+
+    Raises
+    ------
+    TypeError
+        If `dt` is not a ``numpy.dtype``.
+
+    """
+    components = [dt]
+    while 0 != len(components):
+        c = components.pop()
+        if 0 in c.shape:
+            return True
+        if c.names is not None:
+            components.extend([v[0] for v in c.fields.values()])
+        if c.base != c:
+            components.append(c.base)
+    return False
+
+
 def next_unused_name_in_group(grp, length):
     """ Gives a name that isn't used in a Group.
 
@@ -81,7 +279,7 @@ def next_unused_name_in_group(grp, length):
 
 
 def convert_numpy_str_to_uint16(data):
-    """ Converts a numpy.unicode\_ to UTF-16 in numpy.uint16 form.
+    """ Converts a numpy.unicode\\_ to UTF-16 in numpy.uint16 form.
 
     Convert a ``numpy.unicode_`` or an array of them (they are UTF-32
     strings) to UTF-16 in the equivalent array of ``numpy.uint16``. The
@@ -93,7 +291,7 @@ def convert_numpy_str_to_uint16(data):
 
     Parameters
     ----------
-    data : numpy.unicode\_ or numpy.ndarray of numpy.unicode\_
+    data : numpy.unicode\\_ or numpy.ndarray of numpy.unicode\\_
         The string or array of them to convert.
 
     Returns
@@ -132,11 +330,11 @@ def convert_numpy_str_to_uint16(data):
     shape = list(cdata.shape)
     shape[-1] *= (cdata.dtype.itemsize // 2)
     return np.ndarray(shape=shape, dtype='uint16',
-                      buffer=cdata.tostring())
+                      buffer=numpy_to_bytes(cdata))
 
 
 def convert_numpy_str_to_uint32(data):
-    """ Converts a numpy.str\_ to its numpy.uint32 representation.
+    """ Converts a numpy.str\\_ to its numpy.uint32 representation.
 
     Convert a ``numpy.str`` or an array of them (they are UTF-32
     strings) into the equivalent array of ``numpy.uint32`` that is byte
@@ -146,7 +344,7 @@ def convert_numpy_str_to_uint32(data):
 
     Parameters
     ----------
-    data : numpy.str\_ or numpy.ndarray of numpy.str\_
+    data : numpy.str\\_ or numpy.ndarray of numpy.str\\_
         The string or array of them to convert.
 
     Returns
@@ -212,16 +410,16 @@ def convert_to_str(data):
     if isinstance(data, (np.ndarray, np.uint8, np.uint16, np.uint32,
                   np.bytes_, np.unicode_)):
         if data.dtype.name == 'uint8':
-            return data.flatten().tostring().decode('UTF-8')
+            return numpy_to_bytes(data.flatten()).decode('UTF-8')
         elif data.dtype.name == 'uint16':
-            return data.flatten().tostring().decode('UTF-16')
+            return numpy_to_bytes(data.flatten()).decode('UTF-16')
         elif data.dtype.name == 'uint32':
-            return data.flatten().tostring().decode('UTF-32')
+            return numpy_to_bytes(data.flatten()).decode('UTF-32')
         elif data.dtype.char == 'S':
             return data.decode('UTF-8')
         else:
             if isinstance(data, np.ndarray):
-                return data.flatten.tostring().decode('UTF-32')
+                return numpy_to_bytes(data.flatten()).decode('UTF-32')
             else:
                 return data.encode('UTF-32').decode('UTF-32')
 
@@ -232,7 +430,7 @@ def convert_to_str(data):
 
 
 def convert_to_numpy_str(data, length=None):
-    """ Decodes data to Numpy unicode string (str\_).
+    """ Decodes data to Numpy unicode string (str\\_).
 
     Decodes `data` to Numpy unicode string (UTF-32), which is
     ``numpy.str_``, or an array of them. If it can't be decoded, it is
@@ -266,7 +464,7 @@ def convert_to_numpy_str(data, length=None):
 
     Returns
     -------
-    numpy.str\_ or numpy.ndarray of numpy.str\_ or data
+    numpy.str\\_ or numpy.ndarray of numpy.str\\_ or data
         If `data` can be decoded into a ``numpy.str_`` or a
         ``numpy.ndarray`` of them, the decoded version is returned.
         Otherwise, `data` is returned unchanged.
@@ -275,7 +473,7 @@ def convert_to_numpy_str(data, length=None):
     --------
     convert_to_str
     convert_to_numpy_bytes
-    numpy.str\_
+    numpy.str_
 
     """
     # The method of conversion depends on its type.
@@ -303,7 +501,7 @@ def convert_to_numpy_str(data, length=None):
         # needs to be have the dtype essentially changed by having its
         # bytes read into ndarray.
         return np.ndarray(shape=tuple(), dtype='U1',
-                          buffer=data.flatten().tostring())[()]
+                          buffer=numpy_to_bytes(data.flatten()))[()]
     elif isinstance(data, np.ndarray) and data.dtype.char == 'S':
         # We just need to convert it elementwise.
         new_data = np.zeros(shape=data.shape,
@@ -356,9 +554,10 @@ def convert_to_numpy_str(data, length=None):
         for i in range(0, new_data.shape[0]):
             chunk = data[(i*length):((i+1)*length)]
             if data.dtype.name == 'uint32':
-                new_data[i] = np.ndarray(shape=tuple(),
-                                         dtype=new_data.dtype,
-                                         buffer=chunk.tostring())[()]
+                new_data[i] = np.ndarray(
+                    shape=tuple(),
+                    dtype=new_data.dtype,
+                    buffer=numpy_to_bytes(chunk))[()]
             else:
                 new_data[i] = np.unicode_(convert_to_str(chunk))
 
@@ -371,7 +570,7 @@ def convert_to_numpy_str(data, length=None):
 
 
 def convert_to_numpy_bytes(data, length=None):
-    """ Decodes data to Numpy UTF-8 econded string (bytes\_).
+    """ Decodes data to Numpy UTF-8 econded string (bytes\\_).
 
     Decodes `data` to a Numpy UTF-8 encoded string, which is
     ``numpy.bytes_``, or an array of them in which case it will be ASCII
@@ -403,7 +602,7 @@ def convert_to_numpy_bytes(data, length=None):
 
     Returns
     -------
-    numpy.bytes\_ or numpy.ndarray of numpy.bytes\_ or data
+    numpy.bytes\\_ or numpy.ndarray of numpy.bytes\\_ or data
         If `data` can be decoded into a ``numpy.bytes_`` or a
         ``numpy.ndarray`` of them, the decoded version is returned.
         Otherwise, `data` is returned unchanged.
@@ -412,7 +611,7 @@ def convert_to_numpy_bytes(data, length=None):
     --------
     convert_to_str
     convert_to_numpy_str
-    numpy.bytes\_
+    numpy.bytes_
 
     """
     # The method of conversion depends on its type.
@@ -438,7 +637,7 @@ def convert_to_numpy_bytes(data, length=None):
         # needs to be have the dtype essentially changed by having its
         # bytes read into ndarray.
         return np.ndarray(shape=tuple(), dtype='S1',
-                          buffer=data.flatten().tostring())[()]
+                          buffer=numpy_to_bytes(data.flatten()))[()]
     elif isinstance(data, np.ndarray) and data.dtype.char == 'U':
         # We just need to convert it elementwise.
         new_data = np.zeros(shape=data.shape,
@@ -496,9 +695,10 @@ def convert_to_numpy_bytes(data, length=None):
         for i in range(0, new_data.shape[0]):
             chunk = data[(i*length2):((i+1)*length2)]
             if data.dtype.name == 'uint8':
-                new_data[i] = np.ndarray(shape=tuple(),
-                                         dtype=new_data.dtype,
-                                         buffer=chunk.tostring())[()]
+                new_data[i] = np.ndarray(
+                    shape=tuple(),
+                    dtype=new_data.dtype,
+                    buffer=numpy_to_bytes(chunk))[()]
             else:
                 new_data[i] = np.bytes_( \
                     convert_to_str(chunk).encode('UTF-8'))
@@ -748,6 +948,10 @@ def set_attribute(target, name, value):
     """
     if name not in target.attrs:
         target.attrs.create(name, value)
+    elif name == 'MATLAB_fields':
+        if not np.array_equal(value, read_matlab_fields_attribute(
+                target.attrs)):
+            target.attrs.create(name, value)
     elif target.attrs[name].dtype != value.dtype \
             or target.attrs[name].shape != value.shape:
         target.attrs.create(name, value)
